typedef struct
{
	s8 x;
	s8 y;
} Coord_s8;

typedef struct
{
	s32 cell_count;
	s32 max_cells;
	Coord_s8 *cell;
} CellList_s8;

static __not_inline int GoLGrid_to_cell_list (const GoLGrid *gg, CellList_s8 *cl)
{
	if (cl)
		cl->cell_count = 0;
	
	if (!gg || !gg->grid || !cl || cl->max_cells < 0)
		return ffsc (__func__);
	
	if (gg->pop_x_off <= gg->pop_x_on)
		return TRUE;
	
	s32 col_on = gg->pop_x_on >> 6;
	s32 col_off = (gg->pop_x_off + 63) >> 6;
	
	s32 cell_ix = 0;
	s32 col_ix;
	s32 row_ix;
	
	for (col_ix = col_on; col_ix < col_off; col_ix++)
		for (row_ix = gg->pop_y_on; row_ix < gg->pop_y_off; row_ix++)
		{
			u64 grid_word = gg->grid [(gg->col_offset * (u64) col_ix) + (u64) row_ix];
			while (grid_word != 0)
			{
				s32 first_bit = most_significant_bit_u64 (grid_word);
				grid_word &= ~(((u64) 1u) << first_bit);
				
				s32 cell_x = gg->grid_rect.left_x + (64 * col_ix) + (63 - first_bit);
				s32 cell_y = gg->grid_rect.top_y + row_ix;
				
				if (cell_x < -128 || cell_x > 127 || cell_y < -128 || cell_y > 127 || cell_ix >= cl->max_cells)
				{
					cl->cell_count = 0;
					return FALSE;
				}
				
				cl->cell [cell_ix].x = cell_x;
				cl->cell [cell_ix].y = cell_y;
				cell_ix++;
			}
		}
	
	cl->cell_count = cell_ix;
	return TRUE;
}

static __force_inline int GoLGrid_or_cell_list (GoLGrid *gg, const CellList_s8 *cl, s32 x_offs, s32 y_offs)
{
	if (!gg || !gg->grid || !cl)
		return ffsc (__func__);
	
	int not_clipped = TRUE;
	s32 cell_ix;
	for (cell_ix = 0; cell_ix < cl->cell_count; cell_ix++)
		not_clipped &= GoLGrid_set_cell_on (gg, cl->cell [cell_ix].x + x_offs, cl->cell [cell_ix].y + y_offs);
	
	return not_clipped;
}

static __not_inline int GoLGrid_or_text_pattern (GoLGrid *gg, const char *pattern, int left_x, int top_y)
{
	int not_clipped = TRUE;
	s32 text_ix = 0;
	s32 cur_y = top_y;
	s32 cur_x = left_x;
	
	while (TRUE)
	{
		char c = pattern [text_ix++];
		
		if (c == '\0')
			break;
		else if (c == '\n')
		{
			cur_y++;
			cur_x = left_x;
		}
		else
		{
			if (c == ' ' || c == '.')
				; // Do nothing
			else if (c == '*' || c == '@')
				not_clipped &= GoLGrid_set_cell_on (gg, cur_x, cur_y);
			else
				fprintf (stderr, "Illegal character in %s\n", __func__);
			
			cur_x++;
		}
	}
	
	return not_clipped;
}
